# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""Test the Settings service API.
"""
import json
from pathlib import Path

import json5  # type:ignore
import pytest
from strict_rfc3339 import rfc3339_to_timestamp  # type:ignore
from tornado.httpclient import HTTPClientError

from jupyterlab_server.test_utils import big_unicode_string, expected_http_error, validate_request


async def test_get_settings_overrides_dicts(jp_fetch, labserverapp):
    # Check that values that are dictionaries in overrides.json are
    # merged with the schema.
    id = "@jupyterlab/apputils-extension:themes"
    r = await jp_fetch("lab", "api", "settings", id)
    validate_request(r)
    res = r.body.decode()
    data = json.loads(res)
    assert data["id"] == id
    schema = data["schema"]
    # Check that overrides.json file is respected.
    assert schema["properties"]["codeCellConfig"]["default"]["lineNumbers"] is True
    assert len(schema["properties"]["codeCellConfig"]["default"]) == 15


@pytest.mark.parametrize("ext", ["json", "json5"])
async def test_get_settings_overrides_d_dicts(
    jp_fetch, jp_serverapp, make_labserver_extension_app, ext
):
    # Check that values that are dictionaries in overrides.d/*.json are
    # merged with the schema.
    labserverapp = make_labserver_extension_app()
    labserverapp._link_jupyter_server_extension(jp_serverapp)
    id = "@jupyterlab/apputils-extension:themes"
    overrides_d = Path(labserverapp.app_settings_dir) / "overrides.d"
    overrides_d.mkdir(exist_ok=True, parents=True)
    for i in range(10):
        text = json.dumps({id: {"codeCellConfig": {"cursorBlinkRate": 530 + i}}})
        if ext == "json5":
            text += "\n// a comment"
        (overrides_d / f"foo-{i}.{ext}").write_text(text, encoding="utf-8")
    # Initialize labserverapp only after the overrides were created
    labserverapp.initialize()
    r = await jp_fetch("lab", "api", "settings", id)
    validate_request(r)
    res = r.body.decode()
    data = json.loads(res)
    assert data["id"] == id
    schema = data["schema"]
    # Check that the last overrides.d/*.json file is respected.
    assert schema["properties"]["codeCellConfig"]["default"]["cursorBlinkRate"] == 539


async def test_get_settings(jp_fetch, labserverapp):
    id = "@jupyterlab/apputils-extension:themes"
    r = await jp_fetch("lab", "api", "settings", id)
    validate_request(r)
    res = r.body.decode()
    data = json.loads(res)
    assert data["id"] == id
    schema = data["schema"]
    # Check that overrides.json file is respected.
    assert schema["properties"]["theme"]["default"] == "JupyterLab Dark"
    assert "raw" in res


async def test_get_federated(jp_fetch, labserverapp):
    id = "@jupyterlab/apputils-extension-federated:themes"
    r = await jp_fetch("lab", "api", "settings", id)
    validate_request(r)
    res = r.body.decode()
    assert "raw" in res


async def test_get_bad(jp_fetch, labserverapp):
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("foo")
    assert expected_http_error(e, 404)


async def test_listing(jp_fetch, labserverapp):
    ids = [
        "@jupyterlab/apputils-extension:themes",
        "@jupyterlab/apputils-extension-federated:themes",
        "@jupyterlab/codemirror-extension:commands",
        "@jupyterlab/codemirror-extension-federated:commands",
        "@jupyterlab/shortcuts-extension:plugin",
        "@jupyterlab/translation-extension:plugin",
        "@jupyterlab/unicode-extension:plugin",
    ]
    versions = ["N/A", "N/A", "test-version"]
    r = await jp_fetch("lab", "api", "settings/")
    validate_request(r)
    res = r.body.decode()
    response = json.loads(res)
    response_ids = [item["id"] for item in response["settings"]]
    response_schemas = [item["schema"] for item in response["settings"]]
    response_versions = [item["version"] for item in response["settings"]]
    assert set(response_ids) == set(ids)
    assert all(response_schemas)
    assert set(response_versions) == set(versions)
    last_modifieds = [item["last_modified"] for item in response["settings"]]
    createds = [item["created"] for item in response["settings"]]
    assert {None} == set(last_modifieds + createds)


async def test_listing_ids(jp_fetch, labserverapp):
    ids = [
        "@jupyterlab/apputils-extension:themes",
        "@jupyterlab/apputils-extension-federated:themes",
        "@jupyterlab/codemirror-extension:commands",
        "@jupyterlab/codemirror-extension-federated:commands",
        "@jupyterlab/shortcuts-extension:plugin",
        "@jupyterlab/translation-extension:plugin",
        "@jupyterlab/unicode-extension:plugin",
    ]
    r = await jp_fetch("lab", "api", "settings/", params={"ids_only": "true"})
    validate_request(r)
    res = r.body.decode()
    response = json.loads(res)
    response_ids = [item["id"] for item in response["settings"]]
    # Checks the IDs list is correct
    assert set(response_ids) == set(ids)

    # Checks there is only the 'id' key in each item
    assert all(
        (len(item.keys()) == 1 and next(iter(item.keys())) == "id") for item in response["settings"]
    )


async def test_patch(jp_fetch, labserverapp):
    id = "@jupyterlab/shortcuts-extension:plugin"

    r = await jp_fetch(
        "lab", "api", "settings", id, method="PUT", body=json.dumps(dict(raw=json5.dumps(dict())))
    )
    validate_request(r)

    r = await jp_fetch(
        "lab",
        "api",
        "settings",
        id,
        method="GET",
    )
    validate_request(r)
    data = json.loads(r.body.decode())
    first_created = rfc3339_to_timestamp(data["created"])
    first_modified = rfc3339_to_timestamp(data["last_modified"])

    r = await jp_fetch(
        "lab", "api", "settings", id, method="PUT", body=json.dumps(dict(raw=json5.dumps(dict())))
    )
    validate_request(r)

    r = await jp_fetch(
        "lab",
        "api",
        "settings",
        id,
        method="GET",
    )
    validate_request(r)
    data = json.loads(r.body.decode())
    second_created = rfc3339_to_timestamp(data["created"])
    second_modified = rfc3339_to_timestamp(data["last_modified"])

    assert first_created <= second_created
    assert first_modified < second_modified

    r = await jp_fetch(
        "lab",
        "api",
        "settings/",
        method="GET",
    )
    validate_request(r)
    data = json.loads(r.body.decode())
    listing = data["settings"]
    list_data = next(item for item in listing if item["id"] == id)
    # TODO(@echarles) Check this...


#    assert list_data['created'] == data['created']
#    assert list_data['last_modified'] == data['last_modified']


async def test_patch_unicode(jp_fetch, labserverapp):
    id = "@jupyterlab/unicode-extension:plugin"
    settings = dict(comment=big_unicode_string[::-1])
    payload = dict(raw=json5.dumps(settings))

    r = await jp_fetch("lab", "api", "settings", id, method="PUT", body=json.dumps(payload))
    validate_request(r)

    r = await jp_fetch(
        "lab",
        "api",
        "settings",
        id,
        method="GET",
    )
    validate_request(r)
    data = json.loads(r.body.decode())
    assert data["settings"]["comment"] == big_unicode_string[::-1]


async def test_patch_wrong_id(jp_fetch, labserverapp):
    with pytest.raises(HTTPClientError) as e:
        await jp_fetch("foo", method="PUT", body=json.dumps(dict(raw=json5.dumps(dict()))))
    assert expected_http_error(e, 404)


async def test_patch_bad_data(jp_fetch, labserverapp):
    with pytest.raises(HTTPClientError) as e:
        settings = dict(keyMap=10)
        payload = dict(raw=json5.dumps(settings))
        await jp_fetch("foo", method="PUT", body=json.dumps(payload))
    assert expected_http_error(e, 404)


async def test_patch_invalid_payload_format(jp_fetch, labserverapp):
    id = "@jupyterlab/apputils-extension:themes"

    with pytest.raises(HTTPClientError) as e:
        settings = dict(keyMap=10)
        payload = dict(foo=json5.dumps(settings))
        await jp_fetch("lab", "api", "settings", id, method="PUT", body=json.dumps(payload))
    assert expected_http_error(e, 400)


async def test_patch_invalid_json(jp_fetch, labserverapp):
    id = "@jupyterlab/apputils-extension:themes"

    with pytest.raises(HTTPClientError) as e:
        payload_str = "eh"
        await jp_fetch("lab", "api", "settings", id, method="PUT", body=json.dumps(payload_str))
    assert expected_http_error(e, 400)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os

os.environ["JUPYTER_PLATFORM_DIRS"] = "1"
pytest_plugins = ["jupyterlab_server.pytest_plugin"]



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import sys
import warnings

import pytest

from jupyterlab_server.process import Process, WatchHelper, which
from jupyterlab_server.process_app import ProcessApp


def test_which():
    assert which("jupyter")


async def test_process():
    p = Process([sys.executable, "--version"])
    p.get_log().info("test")
    assert p.wait() == 0

    p = Process([sys.executable, "--version"])
    p.get_log().info("test")
    assert await p.wait_async() == 0
    assert p.terminate() == 0


@pytest.mark.skipif(os.name == "nt", reason="Fails on Windows")
async def test_watch_helper():
    helper = WatchHelper([sys.executable, "-i"], ">>>")
    helper.terminate()
    helper.wait()


def test_process_app():
    class TestApp(ProcessApp):
        name = "tests"

    app = TestApp()
    app.initialize_server([])
    try:
        app.initialize()
        with pytest.raises(SystemExit):
            app.start()
    # Kandle exception on older versions of server.
    except Exception as e:
        # Convert to warning so the test will pass on min version test.
        warnings.warn(str(e), stacklevel=2)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import os

import json5  # type:ignore
import pytest

from jupyterlab_server.config import get_page_config


@pytest.mark.parametrize(
    "lib,extension",
    [
        (json, "json"),
        (json5, "json5"),
    ],
)
def test_get_page_config(tmp_path, lib, extension):
    labext_path = [os.path.join(tmp_path, "ext")]
    settings_path = os.path.join(tmp_path, "settings")
    os.mkdir(settings_path)

    with open(os.path.join(settings_path, f"page_config.{extension}"), "w") as fid:
        data = dict(deferredExtensions=["foo"])
        lib.dump(data, fid)

    static_dir = os.path.join(tmp_path, "static")
    os.mkdir(static_dir)
    with open(os.path.join(static_dir, "package.json"), "w") as fid:
        data2 = dict(jupyterlab=dict(extensionMetadata=dict(foo=dict(disabledExtensions=["bar"]))))
        json.dump(data2, fid)

    config = get_page_config(labext_path, settings_path)
    assert config == {
        "deferredExtensions": ["foo"],
        "federated_extensions": [],
        "disabledExtensions": ["bar"],
    }



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import os
import sys

from jupyterlab_server.workspaces_app import (
    WorkspaceExportApp,
    WorkspaceImportApp,
    WorkspaceListApp,
)


def test_workspace_apps(jp_environ, tmp_path):
    sys.argv = [sys.argv[0]]

    data = {
        "data": {
            "layout-restorer:data": {
                "main": {
                    "dock": {
                        "type": "tab-area",
                        "currentIndex": 1,
                        "widgets": ["notebook:Untitled1.ipynb"],
                    },
                    "current": "notebook:Untitled1.ipynb",
                },
                "down": {"size": 0, "widgets": []},
                "left": {
                    "collapsed": False,
                    "current": "filebrowser",
                    "widgets": [
                        "filebrowser",
                        "running-sessions",
                        "@jupyterlab/toc:plugin",
                        "extensionmanager.main-view",
                    ],
                },
                "right": {
                    "collapsed": True,
                    "widgets": ["jp-property-inspector", "debugger-sidebar"],
                },
                "relativeSizes": [0.17370242214532872, 0.8262975778546713, 0],
            },
            "notebook:Untitled1.ipynb": {
                "data": {"path": "Untitled1.ipynb", "factory": "Notebook"}
            },
        },
        "metadata": {"id": "default"},
    }

    data_file = os.path.join(tmp_path, "test.json")
    with open(data_file, "w") as fid:
        json.dump(data, fid)

    app = WorkspaceImportApp(workspaces_dir=str(tmp_path))
    app.initialize()
    app.extra_args = [data_file]
    app.start()

    app1 = WorkspaceExportApp(workspaces_dir=str(tmp_path))
    app1.initialize()
    app1.start()

    app2 = WorkspaceListApp(workspaces_dir=str(tmp_path))
    app2.initialize()
    app2.start()

    app2.jsonlines = True
    app2.start()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import os.path as osp
import shutil
import sys

HERE = osp.abspath(osp.dirname(__file__))
sys.path.insert(0, osp.join(HERE, "..", ".."))

from jupyterlab_server import LabServerApp, _version  # noqa: E402

# -- Project information -----------------------------------------------------

project = "JupyterLab Server"
copyright = "2021, Project Jupyter"
author = "Project Jupyter"

# The short X.Y version.
version = "%i.%i" % _version.version_info[:2]
# The full version, including alpha/beta/rc tags.
release = _version.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "autodoc_traits",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinxcontrib.openapi",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
]

try:
    import enchant  # noqa: F401

    extensions += ["sphinxcontrib.spelling"]
except ImportError:
    pass


myst_enable_extensions = ["html_image"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#
html_theme = "pydata_sphinx_theme"

# Add an Edit this Page button
html_theme_options = {
    "use_edit_page_button": True,
    "navigation_with_keys": False,
}

# Output for github to be used in links
html_context = {
    "github_user": "jupyterlab",  # Username
    "github_repo": "jupyterlab_server",  # Repo name
    "github_version": "master",  # Version
    "doc_path": "/docs/source/",  # Path in the checkout to the docs root
}

# This option generates errors when methods do not have docstrings,
# so disable
numpydoc_show_class_members = False

config_header = """\
.. _api-full-config:


Config file and command line options
====================================

The JupyterLab Server can be run with a variety of command line arguments.
A list of available options can be found below in the :ref:`options section
<options>`.

Defaults for these options can also be set by creating a file named
``jupyter_jupyterlab_server_config.py`` in your Jupyter folder. The Jupyter
folder is in your home directory, ``~/.jupyter``.

To create a ``jupyter_jupyterlab_server_config.py`` file, with all the defaults
commented out, you can use the following command line::

  $ python -m jupyterlab_server --generate-config


.. _options:

Options
-------

This list of options can be generated by running the following and hitting
enter::

  $ python -m jupyterlab_server --help-all

"""


def setup(app):  # noqa: ARG001
    dest = osp.join(HERE, "changelog.md")
    shutil.copy(osp.join(HERE, "..", "..", "CHANGELOG.md"), dest)
    destination = osp.join(HERE, "api/app-config.rst")
    with open(destination, "w") as f:
        f.write(config_header)
        f.write(LabServerApp().document_config_options())



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from unittest.mock import Mock

from tornado.httpserver import HTTPRequest
from tornado.web import Application

from jupyterlab_server.test_utils import validate_request
from jupyterlab_server.themes_handler import ThemesHandler


async def test_get_theme(jp_fetch, labserverapp):
    r = await jp_fetch("lab", "api", "themes", "@jupyterlab", "foo", "index.css")
    validate_request(r)


def test_themes_handler(tmp_path):
    app = Application()
    request = HTTPRequest(connection=Mock())
    data_path = f"{tmp_path}/test.txt"
    with open(data_path, "w") as fid:
        fid.write("hi")
    handler = ThemesHandler(app, request, path=str(tmp_path))
    handler.absolute_path = data_path
    handler.get_content_size()
    handler.get_content("test.txt")

    css_path = f"{tmp_path}/test.css"
    with open(css_path, "w") as fid:
        fid.write("url('./foo.css')")
    handler.absolute_path = css_path
    handler.path = "/"
    handler.themes_url = "foo"
    content = handler.get_content(css_path)
    assert content == b"url('foo/./foo.css')"



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

"""Basic tests for the lab handlers.
"""

import json
import re
from pathlib import Path

import pytest
import tornado.httpclient

from jupyterlab_server.test_utils import expected_http_error


@pytest.fixture
def notebooks(jp_create_notebook, labserverapp):
    nbpaths = (
        "notebook1.ipynb",
        "jlab_test_notebooks/notebook2.ipynb",
        "jlab_test_notebooks/level2/notebook3.ipynb",
    )
    for nb in nbpaths:
        jp_create_notebook(nb)
    return nbpaths


def extract_page_config(html):
    return json.loads(
        re.search(
            r'<script id="jupyter-config-data" type="application/json">\s*(?P<data>.*?)\s*</script>',
            html,
        ).group(  # type: ignore
            "data"
        )
    )


async def test_lab_handler(notebooks, jp_fetch):
    r = await jp_fetch("lab", "jlab_test_notebooks")
    assert r.code == 200
    # Check that the lab template is loaded
    html = r.body.decode()
    assert "Files" in html
    assert "JupyterLab Server Application" in html


async def test_page_config(labserverapp, jp_fetch):
    r = await jp_fetch("lab")
    assert r.code == 200
    # Check that the lab template is loaded
    html = r.body.decode()
    page_config = extract_page_config(html)
    assert not page_config["treePath"]
    assert page_config["preferredPath"] == "/"

    def ispath(p):
        return p.endswith(("Dir", "Path")) or p == "serverRoot"

    nondirs = {k: v for k, v in page_config.items() if not ispath(k)}
    assert nondirs == {
        "appName": "JupyterLab Server Application",
        "appNamespace": "jupyterlab_server",
        "appUrl": "/lab",
        "appVersion": "",
        "baseUrl": "/a%40b/",
        "cacheFiles": True,
        "disabledExtensions": [],
        "federated_extensions": [],
        "fullAppUrl": "/a%40b/lab",
        "fullLabextensionsUrl": "/a%40b/lab/extensions",
        "fullLicensesUrl": "/a%40b/lab/api/licenses",
        "fullListingsUrl": "/a%40b/lab/api/listings",
        "fullMathjaxUrl": "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.7/MathJax.js",
        "fullSettingsUrl": "/a%40b/lab/api/settings",
        "fullStaticUrl": "/a%40b/static/jupyterlab_server",
        "fullThemesUrl": "/a%40b/lab/api/themes",
        "fullTranslationsApiUrl": "/a%40b/lab/api/translations",
        "fullTreeUrl": "/a%40b/lab/tree",
        "fullWorkspacesApiUrl": "/a%40b/lab/api/workspaces",
        "ignorePlugins": [],
        "labextensionsUrl": "/lab/extensions",
        "licensesUrl": "/lab/api/licenses",
        "listingsUrl": "/lab/api/listings",
        "mathjaxConfig": "TeX-AMS_HTML-full,Safe",
        "mode": "multiple-document",
        "notebookStartsKernel": True,
        "settingsUrl": "/lab/api/settings",
        "store_id": 0,
        "terminalsAvailable": True,
        "themesUrl": "/lab/api/themes",
        "translationsApiUrl": "/lab/api/translations",
        "treeUrl": "/lab/tree",
        "workspace": "default",
        "workspacesApiUrl": "/lab/api/workspaces",
        "wsUrl": "",
    }


@pytest.fixture
def serverapp_preferred_dir(jp_server_config, jp_root_dir):
    preferred_dir = Path(jp_root_dir, "my", "preferred_dir")
    preferred_dir.mkdir(parents=True, exist_ok=True)
    jp_server_config.ServerApp.preferred_dir = str(preferred_dir)
    return preferred_dir


async def test_app_preferred_dir(serverapp_preferred_dir, labserverapp, jp_fetch):
    r = await jp_fetch("lab")
    assert r.code == 200
    # Check that the lab template is loaded
    html = r.body.decode()
    page_config = extract_page_config(html)
    api_path = str(serverapp_preferred_dir.relative_to(labserverapp.serverapp.root_dir).as_posix())
    assert page_config["preferredPath"] == api_path


async def test_contents_manager_preferred_dir(jp_root_dir, labserverapp, jp_fetch):
    preferred_dir = Path(jp_root_dir, "my", "preferred_dir")
    preferred_dir.mkdir(parents=True, exist_ok=True)
    try:
        _ = labserverapp.serverapp.contents_manager.preferred_dir
        labserverapp.serverapp.contents_manager.preferred_dir = str(preferred_dir)
    except AttributeError:
        pytest.skip("Skipping contents manager test, trait not present")

    r = await jp_fetch("lab")
    assert r.code == 200
    # Check that the lab template is loaded
    html = r.body.decode()
    page_config = extract_page_config(html)
    api_path = str(preferred_dir.relative_to(labserverapp.serverapp.root_dir).as_posix())
    assert page_config["preferredPath"] == api_path


async def test_notebook_handler(notebooks, jp_fetch):
    for nbpath in notebooks:
        r = await jp_fetch("lab", nbpath)
        assert r.code == 200
        # Check that the lab template is loaded
        html = r.body.decode()
        assert "JupyterLab Server Application" in html


async def test_404(notebooks, jp_fetch):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch("foo")
    assert expected_http_error(e, 404)
